/* 
    Neural Network Simulator

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    (C) 2006 Jason Hunt
    nulluser@gmail.gom 
*/

#include <math.h>
#include <stdio.h>

#include "utility.h"
#include "console.h"
#include "animal.h"
#include "world.h"

extern world_type world;
extern bool mode_manual;
extern bool show_vision;

double max_animal_speed = 0.25;  // The animals cal move no faster than this


/* Crazy GDI code to draw a vector graphics arrow */
void draw_arrow(HDC dc, double x, double y, double dir, double size)
{
    double arrow_base_x = x - size * cos(dir);
    double arrow_base_y = y - size  * sin(dir);
    
    MoveToEx(dc, get_screen_x(arrow_base_x), get_screen_y(arrow_base_y), NULL);
    
    double arrow_tip_x = x + size * cos(dir);
    double arrow_tip_y = y + size * sin(dir);
        
    LineTo(dc, get_screen_x(arrow_tip_x), get_screen_y(arrow_tip_y));    

    double arrow_left_x = x + 0.75 * size * cos(dir - M_PI / 4.0);
    double arrow_left_y = y + 0.75 * size * sin(dir - M_PI / 4.0);

    MoveToEx(dc, get_screen_x(arrow_left_x), get_screen_y(arrow_left_y), NULL);

    LineTo(dc, get_screen_x(arrow_tip_x), get_screen_y(arrow_tip_y));    

    double arrow_right_x = x + 0.75 * size * cos(dir + M_PI / 4.0);
    double arrow_right_y = y + 0.75 * size * sin(dir + M_PI / 4.0);

    LineTo(dc, get_screen_x(arrow_right_x), get_screen_y(arrow_right_y));    
}
/* End of draw arrow */


/* Draw a sensor for an animal */
void draw_sensor(HDC dc, double x, double y, double dir, double dist, double size, COLORREF c)
{
    double sx = x + dist * cos(dir);
    double sy = y + dist * sin(dir);
    
    RECT r;
    
    r.left = get_screen_x(sx - size / 2.0);
    r.top = get_screen_y(sy - size / 2.0);
    r.right = get_screen_x(sx + size / 2.0);
    r.bottom = get_screen_y(sy + size / 2.0);
    
    HBRUSH b = CreateSolidBrush(c);

    HBRUSH hold = (HBRUSH) SelectObject(dc, b);

    Ellipse(dc, r.left, r.top, r.right, r.bottom);    
    
    SelectObject(dc, hold);    
        
    DeleteObject(b);
}
/* End of draw sensor */


/* Decode the sensor into a color */
COLORREF get_sensor_color( sensor_type &s )
{
    if (s.wall > 0) return(RGB(255, 0, 0));     // Red for well
    if (s.food > 0) return(RGB(0, 255, 0));     // Gree for food
    if (s.other > 0) return(RGB(0, 0, 255));    // Blue for other

    return(RGB(255, 255, 255));                 // No hit, return white
}
/* End of get_sensor_color */


/* Draw an animal to a device context */
void animal_draw(HDC dc, animal_type *a)
{
    if (a == NULL) return;

    RECT r;
    get_bounding_rect(a->x, a->y, a->x_size, a->y_size, r);

    // Draw root animal in green
    if (a == world.animals)
        SelectObject(dc, (HBRUSH)world.green_brush);

    Ellipse(dc, r.left, r.top, r.right, r.bottom);

    SelectObject(dc, (HBRUSH)world.white_brush);

    // Get distance from center and size of the sensors
    double size = a->sensor_size;
    double dist = a->sensor_dist;

    // Draw the direction arrow
    draw_arrow(dc, a->x, a->y, a->dir, size);
 
    // Draw each sensor
    // The sensor index increases in a clockwise direction

    double angle = - M_PI / 3.0;
    double delta = 2 * M_PI / 6.0;

    for (int i = 0; i < 6; i++, angle += delta)
        draw_sensor(dc,  a->x, a->y, a->dir + angle, dist, size, 
                       get_sensor_color(a->touch[i]));
                 
                       
    if (show_vision)                       
    for (int i = 0; i < a->num_vision_points * 2; i++)
    {
        if (a->vision[i/ 2].food > 0)
        {

            MoveToEx(dc, get_screen_x(a->x),get_screen_y(a->y),NULL);
            LineTo(dc, get_screen_x(a->x+a->end_x[i]),get_screen_y(a->y+a->end_y[i]));
        }
    }
                       
                       
}
/* End of draw sensors */



/* Check the spac that the sensor tip occupies, and set values to match */
void check_sensor(double x, double y, double dir, double dist, 
                    sensor_type &s)
{
    // Determine location of sensor tip
    double sx = x + dist * cos(dir);
    double sy = y + dist * sin(dir);
    
    s.wall = -0.5;

    if (sx < 0 || sx > world.x_size) s.wall =0.5;
    if (sy < 0 || sy > world.y_size) s.wall = 0.5;
    
    // Check collosions with all other animals    
    animal_type *a = world.animals;

    s.other = -0.5;
        
    while (a != NULL && s.other == -0.5)
    {
        double dist = (sx - a->x) * (sx - a->x) + (sy - a->y) * (sy - a->y);

        // compute radius of amimal squared
        double radius = (a->x_size + a->y_size) * (a->x_size + a->y_size) / 16.0;

        if ( dist  < radius)
            s.other = 0.5;
                
        a = a->next;
    }    
    
    s.food = -0.5;

    plant_type *p = world.plants;
        
    while (p != NULL && s.food == -0.5)
    {
        double dist = (sx - p->x) * (sx - p->x) + (sy - p->y) * (sy - p->y);

        // compute radius of amimal squared
        double radius = (p->x_size + p->y_size) * (p->x_size + p->y_size) / 16.0;

        if ( dist  < radius)
            s.food = 0.5;
                
        p = p->next;
    } 
}
/* End of check sensor */




/* Check the spac that the sensor tip occupies, and set values to match */
void check_vision(double x, double y, double dir, int vision_points, sensor_type s[3], double end_x[3], double end_y[3])
{              
    for (int i = 0; i < vision_points*2; i++)
    {
        end_x[i] = x;
        end_y[i] = y;
    }
    
    double left_bounds[vision_points];
    double right_bounds[vision_points];
    double hit_counts[vision_points];
    

    double vision_width = M_PI/2.0;      // Width of entire vision cone
    double length = 0.55;    
    
    double cone_width = vision_width / (double)vision_points;
    
    double cur_dir = dir - vision_width / 2.0;
    
    for (int i = 0; i < vision_points; i++)
    {
        left_bounds[i] = angle_wrap(cur_dir);
        right_bounds[i]= angle_wrap(cur_dir + cone_width);
        
        end_x[2*i] = length * cos(left_bounds[i]);
        end_y[2*i] = length * sin(left_bounds[i]);        
        
        end_x[2*i+1] = length * cos(right_bounds[i]);
        end_y[2*i+1] = length * sin(right_bounds[i]);        
        
        cur_dir += cone_width;
        
        hit_counts[i] = 0;
    }
    
    plant_type *p = world.plants;
    
    // Check the plants for line of sight counts        
    while (p != NULL)
    {
        double p_angle = angle_wrap(atan2(p->y - y, p->x - x));
    
        for (int i = 0; i < vision_points; i++)
        {
            if (p_angle > left_bounds[i] && p_angle < right_bounds[i])
                hit_counts[i]++;
        }    
        
        p = p->next;       
    }
 

    // Update vision values if there are any plants in sight
    for (int i = 0; i < vision_points; i++)
    {
        if (hit_counts[i] > 0)
            s[i].food = hit_counts[i] * 0.1;  
        else
            s[i].food = -0.5;    
    }
}
/* End of check vision */



/* See if the animal is eating anything */
void animal_check_plants( animal_type *a )
{
    plant_type *prev, *p = world.plants;

    prev = NULL;

    while(p != NULL)
    {
        // Calculate distance between centers
        double dist = (a->x - p->x) * (a->x - p->x) + (a->y - p->y) * (a->y - p->y);

        // compute radius of amimal squared
        double radius = (a->x_size + a->y_size) * (a->x_size + a->y_size) / 8.0 + 
                        (p->x_size + p->y_size) * (p->x_size + p->y_size) / 8.0; 

        if (dist  < radius)
        {
            a->food_capture++;         // Increase score
    
            remove_plant(p);    // Remove plant 

            p = prev;             // Point to previous plant

        }  else
        { // advance to the next plant

            prev = p;
            p=p->next;
        }
    }
}
/* End of animal_check_plants */



/* Update all values for the aminal */
void animal_update(animal_type *a)
{
    if (a == NULL) return;

    // Check for the animal overlapping a plant
    animal_check_plants(a);

    // Read all sensors    
    
    // Claculate distance to tip of sensor
    double dist = a->sensor_dist + a->sensor_size / 2.0;    

    // Calculate angle and delta for sensor check
    double angle = - M_PI / 3.0;
    double delta = 2 * M_PI / 6.0;

    // Check the sensors moving clockwise around the animal
    for (int i = 0; i < 6; i++, angle += delta)
        check_sensor(a->x, a->y, a->dir + angle, dist,a->touch[i]);


    angle = - M_PI / 5.0;
    delta = M_PI / 5.0;    

    check_vision(a->x, a->y, a->dir, a->num_vision_points, a->vision, a->end_x, a->end_y);


    // Now that the input is final, we can process the neural network
    process_neurons(a->neurons);

    // We need to extract the output values from the drive neurons
    // into the left and right drive values

    // Make sure the root one is under user control
    if (a != world.animals || !mode_manual)
    {
        // Get left drive neuron
        neuron_type *n = get_neuron(a->neurons, a->l_drive_id);
        
        // Make sure that get_neuron returned something    
        if (n != NULL)
        {
            if (n->out > max_animal_speed ) n->out = max_animal_speed;
            if (n->out < -max_animal_speed ) n->out = -max_animal_speed;
                
            a->l_drive = n->out;
        }
        else 
           a->l_drive = 0;

        // Get right drive neuron
        n = get_neuron(a->neurons, a->r_drive_id);

        if (n != NULL)
        {
            if (n->out > max_animal_speed ) n->out = max_animal_speed;
            if (n->out < -max_animal_speed ) n->out = -max_animal_speed;

            a->r_drive = n->out;
        }            
        else 
            a->r_drive = 0;
    }
    
    // Figure out speed and diretion delta from left and right drives

    // Direction change rate is the difference of drive speeds
    a->ddir = a->l_drive - a->r_drive;

    // Speed is the sum of drive speeds
    a->speed = (a->l_drive + a->r_drive) / 2.0;    

    // Add delta to direction
    a->dir += a->ddir;

    if (a->dir > 2*M_PI ) a->dir -= 2*M_PI;
    if (a->dir < 0 ) a->dir += 2*M_PI;

    // Calulate x and y deltas
    a->dx = a->speed * cos(a->dir);
    a->dy = a->speed * sin(a->dir);
    
    // Add deltas to positon
    a->x += a->dx;
    a->y += a->dy;
    
    // Wrap around mode
    if (world.wrap_around)
    {
        if (a->x < a->x_size/2)
            a->x = world.x_size - a->x_size/2;
        
        if (a->x > world.x_size - a->x_size/2)
            a->x = a->x_size/2;
        
        if (a->y < a->y_size/2)
            a->x = world.y_size - a->y_size/2;
        
        if (a->y > world.y_size - a->y_size/2)
            a->y = a->y_size/2;
    }
    else    
    {
        // Check collision with outer walls        
        if (a->x < a->x_size/2)
        {            
            a->x = a->x_size/2;
            a->wall_hit++;    
        }
        
        if (a->x > world.x_size - a->x_size/2)
        {
            a->x = world.x_size - a->x_size/2;
            a->wall_hit++;        
        }
            
        if (a->y < a->y_size/2)
        {
            a->y = a->y_size/2;
            a->wall_hit++;                    
        }
            
        if (a->y > world.y_size - a->y_size/2)
        {
            a->y = world.y_size - a->y_size/2;
            a->wall_hit++;        
        }
    }
}


/* Manually Set motions for the first anaimal */
void animal_motion(double l, double r)
{
    if (world.animals == NULL) return;

    world.animals->l_drive = l;
    world.animals->r_drive = r;
}
/* End of animal_motion */


/* Setup inital neurons for an animal */
void animal_setup_neurons(animal_type *a)
{
    // Make sure pointer is NULL
    a->neurons = NULL;
        
    int n = 0;  // Neuron ID index
    
    // Add neurons imput layer neurons
    for (int i = 0; i < a->neuron_width; i++)
    {
        add_neuron(&a->neurons, n, 0., 1., 0., 0., 0., 0.);
                    
        // Connect the neurons to the sensors                    
        // Each neuron will have three sources
        // Wall, food and other    

        // Wall is -4 weight
        
        for (int j = 0; j < a->num_touch_points; j++)     
        {        
            neuron_connect(a->neurons, n, &a->touch[j].wall, 0, 1, 1);

            // Food is 1 weight
            neuron_connect(a->neurons, n, &a->touch[j].food, 0, 1, 1);

            neuron_connect(a->neurons, n, &a->touch[j].other, 0, 1, 1);
        }
        
        for (int j = 0; j < a->num_vision_points; j++)
           neuron_connect(a->neurons, n, &a->vision[j].food, 0, 1, 1);                
        
        n++;
    }
     
    // add the other neuron layers
    for (int i = 0; i < a->neuron_depth - 1; i++)
    {
        for (int j = 0; j < a->neuron_width; j++)
        {
            add_neuron(&a->neurons, n, 0., 1., 0., 0., 0., 0.);
                
           // Connect to previous row
     
            for (int k = 0; k < a->neuron_width ; k++)            
                neuron_connect(a->neurons, n, 
                               &get_neuron(a->neurons, i*a->neuron_width + k)->out, 0, 1, 1);                
            
            n++;            
        }       
    }
          
            
    a->l_drive_id = n++;
    a->r_drive_id = n++;
            
    add_neuron(&a->neurons, a->l_drive_id, 0., 1., 0., 0., 0., 0.);    
    add_neuron(&a->neurons, a->r_drive_id, 0., 1., 0., 0., 0., 0.);


    unsigned int last_row = a->neuron_depth-1;

    // Connect the drive neurons to the last layer
    for (int i = 0; i < a->neuron_width; i++)
    {
        neuron_connect(a->neurons, a->l_drive_id, &get_neuron(a->neurons, i + last_row)->out, 0, 1, 1);        
        neuron_connect(a->neurons, a->r_drive_id, &get_neuron(a->neurons, i + last_row)->out, 0, 1, 1);                
    }


   // Connect the first layer to drive neurons for feed back
    for (int i = 0; i < a->neuron_width; i++)
    {
        neuron_connect(a->neurons, i, &get_neuron(a->neurons, a->l_drive_id)->out, 0, 1, 1);        
        neuron_connect(a->neurons, i, &get_neuron(a->neurons, a->l_drive_id)->out, 0, 1, 1);                
    }

    
    a->num_neurons = n;
    
}
/* End of animal_setup_neurons */


/* Setup inital neurons for an animal */
void animal_setup_neurons_old(animal_type *a)
{
    // Make sure pointer is NULL
    a->neurons = NULL;
    
    // Add neurons connected to touch sensors 
    // Neuron IDs [0..5]
    for (int  i =0; i < 6; i++)
    {
        add_neuron(&a->neurons, i, 0., 1., 0., 0., 0., 0.);
                    
        // Connect the neurons to the sensors                    
        // Each neuron will have three sources
        // Wall, food and other    

        // Wall is -4 weight
        neuron_connect(a->neurons, i, &a->touch[i].wall, 0, -1, 1);

        // Food is 1 weight
        neuron_connect(a->neurons, i, &a->touch[i].food, 0, 1, 1);

        neuron_connect(a->neurons, i, &a->touch[i].other, 0, 0.5, 1);
    }

    // Add first layer of neurons   IDs:  [6..11]
 /*   for (int  i =6; i < 12; i++)
        add_neuron(&a->neurons, i, 0., 1., 0., 0., 0., 0.);

    // Connect first later to the sensor compactor neurons
    for (int  i =6; i < 12; i++)
        for (int  j =0; j < 6; j++)    
            neuron_connect(a->neurons, i, &get_neuron(a->neurons, j)->out, 0, 1, 1);
 
    // Add second layer of neurons   IDs:  [12..17]
    for (int  i = 12; i < 18; i++)
        add_neuron(&a->neurons, i, 0., 1., 0., 0., 0., 0.);

    // Connect first later to the sensor compactor neurons
    for (int  i = 12; i < 18; i++)
        for (int  j =6; j < 12; j++)    
            neuron_connect(a->neurons, i, &get_neuron(a->neurons, j)->out, 0, 1, 1);
 */
    // Add the drive neurons  [18..19]
    add_neuron(&a->neurons, 6, 0., 1., 0., 0., 0., 0.);    
    add_neuron(&a->neurons, 7, 0., 1., 0., 0., 0., 0.);

    // Connect the drive neurons to the secone layer
  //  for (int  i = 12; i < 18; i++)

    for (int  i = 0; i < 6; i++)
    {
        neuron_connect(a->neurons, 6, &get_neuron(a->neurons, i)->out, 0, 1, 1);        
        neuron_connect(a->neurons, 7, &get_neuron(a->neurons, i)->out, 0, 1, 1);                
    }
        
    // Connect the first layer of neurons to the output of the motor drives for feed back
         
    // Neuron IDs [0..5]
/*    for (int  i =0; i < 6; i++)
    {
        neuron_connect(a->neurons, i, &get_neuron(a->neurons, 18)->out, 0, 0.001, 1);
        neuron_connect(a->neurons, i, &get_neuron(a->neurons, 19)->out, 0, 0.001, 1);
    }*/
    
    // Start moving at random
    get_neuron(a->neurons, 6)->out = 1;
    get_neuron(a->neurons, 7)->out = 1;

    get_neuron(a->neurons, 6)->bias = 0.1;
    get_neuron(a->neurons, 7)->bias = 0.1;

    get_neuron(a->neurons, 6)->last_weight = 0.5;
    get_neuron(a->neurons, 7)->last_weight = 0.5;


    // Assign the IDs of the drive neurons
    a->l_drive_id = 6;
    a->r_drive_id = 7;

    // Display for debugging
 //   animal_show_neurons(a);              
}
/* End of animal_setup_neurons */
/* Display entire neuron list */
void animal_show_neurons(animal_type *a)
{
    neuron_type *n = a->neurons;
    
    while (n != NULL)
    {  
        show_neuron(a->neurons, n->id);        
        n = n->next;
    }
}
/* End of animal_show_neurons */


/* Randomize an animal, normally called when the first generation is made */
void animal_randomize( animal_type *a)
{    
    unsigned int size = animal_get_chromosome_size(a);
   
    double c[size];                                 // Chromosome data
    
    animal_export_chromosome(a, c);                 // Export

// before change, 188 gen 50 terminal
// after 154 gen 50 , terminal: 173 (200)
    for (int i = 0; i < size; i++)
        c[i] = f_rand(-1, 1);

//    animal_mutate_chromosome(a, c, size, 1, 2);     // Mutate severely

    animal_inport_chromosome(a, c);                 // Import back into animal        
}
/* End of animal_randomize */



void animal_reset( animal_type *a )
{
    a->x = f_rand(world.animals->x_size, world.x_size - world.animals->x_size);
    a->y = f_rand(world.animals->y_size, world.y_size - world.animals->y_size);

    a->dir = f_rand(0, 2*M_PI);

    a->r_drive = 0;
    a->l_drive = 0;
                    
    a->food_capture = 0;                
    a->wall_hit = 0; 
                
                                
    // Clear touch sensor inputs
    for (int i = 0; i < 6; i++)    
    {
        a->touch[i].wall = 0;
        a->touch[i].food = 0;                
        a->touch[i].other = 0;
    }    

    // Clear vision inputs    
    for (int i = 0; i < a->num_vision_points; i++)    
    {
      //  a->touch[i].wall = 0;
      //  a->touch[i].food = 0;                
      //  a->touch[i].other = 0;
        

        // Clear vision vectors
        a->end_x[2*i] = a->x;
        a->end_y[2*i] = a->y;
        
        a->end_x[2*i+1] = a->x;
        a->end_y[2*i+1] = a->y;
            
       
    }    
              
}
/* End of animal_reset */

/* Setup an animal for the first time */
void animal_init(animal_type *a)
{
    if (a == NULL) return;        
                
    a->neuron_width = world.animal_neuron_width;
    a->neuron_depth = world.animal_neuron_depth;

    a->num_vision_points = 5;
    a->num_touch_points = 6;
    
    a->num_neurons = 0;
         
    a->sensor_dist = (a->x_size + a->y_size) / 4.0;
    a->sensor_size = (a->x_size + a->y_size) / 8.0;    
                                
    // These should be set correctly by the neuron init
    a->l_drive_id = 0;
    a->r_drive_id = 0;

    // Create neural network for this animal
    animal_setup_neurons(a);

    animal_reset(a);                    // Reset position and motion
    
    animal_randomize(a);                // Mutate randomly
}
/* End of animat_init */



/* Export the neural network data into a chromosome */
void animal_inport_chromosome(animal_type *a, double *c)
{
    neuron_type *n = a->neurons;

    while (n != NULL)
    {
        // add sources
        for (int i = 0; i < n->num_sources; i++)
        {
//            n->source_bias[i] = *(c++);
            n->source_scaler[i] = *(c++);
            //n->source_power[i] = (int)*(c++);
        }
        n->bias = *(c++);
//        n->sum_weight = *(c++);
//        n->prod_weight = *(c++);
//        n->intr_weight = *(c++);
//        n->derv_weight = *(c++);

//        n->feedback_weight = *(c++);
        
        n = n->next;    
    }
}
/* End of animal_inport_chromosome */


/* Export the neural network data into a chromosome */
void animal_export_chromosome(animal_type *a, double *c)
{
    neuron_type *n = a->neurons;

    while (n != NULL)
    {
        // add sources
        for (int i = 0; i < n->num_sources; i++)
        {
          //  *(c++) = n->source_bias[i];
            *(c++) = n->source_scaler[i];
          //  *(c++) = (double)n->source_power[i];
        }
         *(c++) = n->bias;
      //   *(c++) = n->sum_weight;
//         *(c++) = n->prod_weight;
//         *(c++) = n->intr_weight;
//         *(c++) = n->derv_weight;
//         *(c++) = n->feedback_weight;

        n = n->next;    
    }
}
/* End of animal_export_chromosome */


/* Preform a mutation on a chomosome */
void animal_mutate_chromosome(animal_type *a, double *c, unsigned int size, 
                              double chance, double rate)
{
    for (int i = 0; i < size; i++)
    {
        if (rand() < chance * RAND_MAX)
            *(c++) += f_rand(-rate, rate);
    }
}
/* End of animal_mutate_chromosome */



/* Calculate the chromosome size */
unsigned int animal_get_chromosome_size(animal_type *a)
{
    neuron_type *n = a->neurons;

    unsigned int s = 0;

    while (n != NULL)
    {
        s += n->num_sources + 1;
        n = n->next;    
    }

    return(s);
}
/* End of animal_get_chromosome_size */







