/* 
    Neural Network Simulator

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    (C) 2006 Jason Hunt
    nulluser@gmail.gom 
*/

#include <windows.h>

#include <stdio.h>
#include <time.h>

#include "resource.h"
#include "main.h"
#include "display.h"
#include "console.h"
#include "world.h"


extern world_type world;


HDC memdc;                               // Device context for offscreen drawing
HBITMAP mem_bitmap;                      // Bitmap for memdc

unsigned int display_x, display_y;       // Current Size of the client area

bool keys[256] = { false };              // Key array, true if pressed

unsigned int updates = 0;                // Number of updates since last second
unsigned int frames = 0;                 // Number of updates since last second

unsigned int updates_per_second = 0;     // Number of updates per second
unsigned int frames_per_second = 0;      // Number of updates per second


bool mode_manual = false;                // True if in manual animal control mode
bool show_vision = false;                // True if animal vision is shown

unsigned int timer_val = 40;             // Inital frame timer value

HWND neuron_window;                      // Handle to neuron window

bool neuron_window_active = false;       // Neuron windows not displayed

animal_type *cur_animal = NULL;          // Pointer to the currently selected animal



/* Init the entire system */
void system_startup( HWND hwnd )
{
    srand(time(NULL));
    
    memdc_setup(hwnd);

    world_setup();

    world_update();

    world_draw(memdc);    

    SetTimer(hwnd, UPDATE_ID, timer_val, NULL);        
    SetTimer(hwnd, SECOND_ID, 1000, NULL);        
}
/* End of system_startup */


/* De-Init the entire system */
void system_shutdown( void )
{
    world_destroy();            
            
    DeleteObject(mem_bitmap);
    DeleteDC(memdc);   
}
/* End of system_shutdown */


/* Confirm program reset */
void reset_check( void )
{
    if (MessageBox(NULL, "Are you sure?", "Reset simulation", MB_OKCANCEL) == IDOK)
        world_reset();                   
}
/* End of reset_check */
    


/* Called when a key is pressed */
void key_down(HWND hwnd, WPARAM w)
{
    // Mark key as pressed
    keys[w & 0xff] = true;  

    // Deal with single press events here
    if (w == 'M') mode_manual = !mode_manual;      // Toggle manual mode

    if (w == 'X') world.graphics_on = !world.graphics_on;      // Toggle manual mode

    if (w == 27) PostQuitMessage(0);

    if (keys[' ']) world_add_animal(); 

    if (keys[VK_DOWN])
    {
        timer_val++;
        
        SetTimer(hwnd, UPDATE_ID, timer_val, NULL);                     
    }
    
    if (keys[VK_UP])
    {
        timer_val --;
        
        if (timer_val < 0) timer_val = 0;
        
       SetTimer(hwnd, UPDATE_ID, timer_val, NULL);
    }
   
    if (keys[VK_LEFT])
    {
        if (world.updates_per_display > 1)
        world.updates_per_display--;
    }   
   
    if (keys[VK_RIGHT])
        world.updates_per_display++;

    if (w == 'R') reset_check();
    
    if (keys['P']) add_random_plant();                  

    if (keys['G']) 
        for (int i =0; i < world.cycles_per_generation; i++)
            world_update();
    
    if (keys['1'])
        world.updates_per_display = 1;
    
    if (keys['2'])
        world.updates_per_display = 200;
        
    if (keys['F'])
        world_free_plants();
    
    if (keys['T']) world.wrap_around = !world.wrap_around;
    
    if (keys['Q'])    
        animal_show_neurons(world.animals);
    
    if (keys['N'])
    {
        cur_animal = world.animals;
        ShowWindow(neuron_window, true);
    }
    
}
/* End of key_down */


/* Deal with main menu commands */
void main_menu(HWND hwnd, unsigned int id)
{
    switch(id)
    {
        case ID_EXIT:           DestroyWindow(hwnd);             return;
        case ID_SAVE:           world_save();                    return;
        case ID_OPEN:           world_load();                    return;            

        case ID_RESET:          reset_check();                   return;                
        case ID_REMOVE_PLANTS:  world_free_plants();             return;   
        case ID_ADD_PLANT:      add_random_plant();              return;   
        case ID_REMOVE_ANIMALS: world_free_animals();            return;   
        case ID_ADD_ANIMAL:     world_add_animal();              return;             

        case ID_NORMAL_SPEED:   world.updates_per_display = 1;   return;
        case ID_MAX_SPEED:      world.updates_per_display = 100; return;
        case ID_PAUSE:          world.running = !world.running;  return;  
        
        case ID_SHOW_VISION:    show_vision = true;              return;
        case ID_HIDE_VISION:    show_vision = false;             return;  
              
    }    
}
/* End of main menu */



/* Called when a key is released */
void key_up(WPARAM w)
{
    keys[w & 0xff] = false;            
}
/* End of key up */


/* Process currently pressed keys */
void check_keys( void )
{
    // Decode WASD keys for manual override

//    write_console("check_keys\n");

    // Forward
    if (keys['W']) 
    {
        if (keys['A']) animal_motion(0.015, 0.03); else     // Forward + left
        if (keys['D']) animal_motion(0.03, 0.015); else     // Forward + right
            animal_motion(0.03, 0.03);                      // Normal forward
                
    } else
    // Reverse
    if (keys['S'])
    {
        if (keys['A']) animal_motion(-0.015, -0.03); else   // Reverse + left
        if (keys['D']) animal_motion(-0.03, -0.015); else   // Reverse + right
            animal_motion(-0.03, -0.03);                    // Normal reverse
        
    } else
    if (keys['A']) animal_motion(-0.03, 0.03); else         // Rotate left
    if (keys['D']) animal_motion(0.03, -0.03); else         // Rotate right
        animal_motion(0.0, 0.0);                            // No key, stop
}
/* End of check keys */


/* Deal with the system timer */
void timer_message(HWND hwnd, WPARAM wparam)
{
    if (wparam == UPDATE_ID)
    {
        check_keys();

        for (int i = 0; i < world.updates_per_display; i++)
            world_update();

        world_draw(memdc);
    
        frames++;

        InvalidateRect(hwnd, NULL, 0);                            
        return;
    }

    if (wparam == SECOND_ID)
    {
        updates_per_second = updates;
        updates = 0;

        frames_per_second = frames;
        frames = 0;

        world_debug_update();      
    }            
}
/* End of timer_message */


/* Deal with paint message */
void window_paint(HWND hwnd)
{
    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(hwnd, &ps);

    char buff[400];
           
    sprintf(buff, "Frame rate: %3d Update Rate  %d ", frames_per_second, updates_per_second);
                       
    TextOut(memdc, 1, 1, buff, strlen(buff)); 

    // Copy memory device context to display 
    BitBlt(hdc, 0, 0, display_x, display_y, memdc, 0, 0, SRCCOPY);
            
    EndPaint(hwnd, &ps); 
    
    InvalidateRect(neuron_window, NULL, 0);      
}
/* Wnd of window_paint */



/* Setup the off screen drawing system */
void memdc_setup( HWND hwnd )
{
    RECT r;
    GetClientRect(hwnd, &r);

    // Update stored client area size
    display_x = r.right;
    display_y = r.bottom;

    HDC hdc = GetDC(hwnd);

    // Create memory device context and bitmap            
    memdc = CreateCompatibleDC(hdc);
    mem_bitmap = CreateCompatibleBitmap(hdc, r.right, r.bottom);
            
    // Select the bitmap into the memort dc
    SelectObject(memdc, mem_bitmap);
            
    ReleaseDC(hwnd, hdc);     
}
/* End of memdc_setup */


/* Chhose the correct brush color for the neuron */
void set_neuron_brush(HDC hdc, double val)
{
    if (val == 0)
        SelectObject(hdc, (HBRUSH)world.white_brush);
    else
    if (val > 0)
        SelectObject(hdc, (HBRUSH)world.blue_brush);
    else
        SelectObject(hdc, (HBRUSH)world.red_brush);     
}
/* End of set_neuron_brush */


/* Display neurons for first animal */
void paint_neuron_window(HWND hwnd)
{
    if (cur_animal == NULL) return;
    
    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(hwnd, &ps);
    SelectObject(hdc, (HBRUSH)world.white_brush);

    RECT r;
    
    GetClientRect(hwnd, &r);
    
    Rectangle(hdc, r.left, r.top, r.right, r.bottom);
    
    neuron_type *n;;
        
    int index = 0;
                        
    int x_size = r.right / cur_animal->neuron_width;
    int y_size = r.bottom / (cur_animal->neuron_depth + 1);
    
                    
    int y = 0, x = 0;
        
    while ((n = get_neuron(world.animals->neurons, index)) != NULL)
    {
        set_neuron_brush(hdc, n->out);                    
        Ellipse(hdc, x*x_size, y*y_size, (x+1)*x_size, (y+1)*y_size);         
           
        // Next line of neurons     
        if (++x == cur_animal->neuron_width ) { y++; x = 0; }
        
        index++;
    }            
                                         
    EndPaint(hwnd, &ps);    
}
/* End of paint_neuron_window */



/* Window function */
LRESULT CALLBACK neuron_win_proc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    switch(message)
    {
        case WM_PAINT: 
            paint_neuron_window(hwnd);
            return(0);
                
        case WM_CHAR:
            if (wparam == 'n')
                ShowWindow(neuron_window, false);                  
            break;        
        
        default:
            return(DefWindowProc(hwnd, message, wparam, lparam));    
    }
    return(0);
}
/* Wnd of window function */



/* Window function */
LRESULT CALLBACK main_win_proc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    switch(message)
    {
        case WM_TIMER: timer_message(hwnd, wparam); return(0);

        case WM_CREATE: system_startup(hwnd); return(0);

        case WM_PAINT: window_paint(hwnd); return(0);

        case WM_DESTROY: 
            system_shutdown();            
            PostQuitMessage(0);
            return(0);
                                        
        case WM_SIZE:               // Process size message
            // These are alreadt setup, delete them
            DeleteObject(mem_bitmap);
            DeleteDC(memdc);
            
            // Setup again with the new size
            memdc_setup(hwnd);
            return(0);
                
        case WM_KEYDOWN:            // Process keydown
            key_down(hwnd, wparam);
            return(0);
    
        case WM_KEYUP:              // Process keyup
            key_up(wparam);
            return(0);
                
                
        case WM_COMMAND:
            main_menu(hwnd, LOWORD(wparam));
            return(0);
                
        default:
            return(DefWindowProc(hwnd, message, wparam, lparam));    
    }
    return(0);
}
/* Wnd of window function */



/* Create the neuron activity display window */
void create_neuron_window( HINSTANCE hinst )
{
    char name[] = "Neuron";
    
    WNDCLASS wc;
    
    wc.hInstance = hinst;
    wc.lpszClassName = name;
    wc.lpfnWndProc = neuron_win_proc;
    wc.style = 0;
    
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.lpszMenuName = NULL;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    
    wc.hbrBackground = NULL;
    
    if (!RegisterClass(&wc)) return;
    
    neuron_window = CreateWindow(name, name, WS_BORDER, 
                            640, 120, // x, y
                            200, 140,  // width, height
                            HWND_DESKTOP, NULL, hinst, NULL);
                            
    ShowWindow( neuron_window, neuron_window_active);
    UpdateWindow( neuron_window );   
}
/* End of create_neuron_window */


/* Winmain */
int WINAPI WinMain(HINSTANCE hinst, HINSTANCE pinst, LPSTR args, int mode)
{
    char name[] = "Neural Network Evolution System";
    
    WNDCLASS wc;
    
    wc.hInstance = hinst;
    wc.lpszClassName = name;
    wc.lpfnWndProc = main_win_proc;
    wc.style = 0;
    
    wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.lpszMenuName = "MAINMENU";
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    
    wc.hbrBackground = NULL;
    
    if (!RegisterClass(&wc)) return(0);
    
    HWND hwnd = CreateWindow(name, name, WS_OVERLAPPEDWINDOW, 
                            120, 120, // x, y
                            640,480, // height, width
                            HWND_DESKTOP, NULL, hinst, NULL);
                            
    ShowWindow( hwnd, mode);
    UpdateWindow( hwnd );

    create_neuron_window(hinst);

    MSG msg;

    while (GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
        
    return (msg.wParam);
}
/* End of winmain */









