/* 
    Neural Network Simulator

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    (C) 2006 Jason Hunt
    nulluser@gmail.gom 
*/

#include <windows.h>
#include <math.h>
#include <stdio.h>

#include "utility.h"
#include "console.h"
#include "animal.h"
#include "world.h"


extern int display_x;   // Size of client area in pixels
extern int display_y;
extern unsigned int updates;

world_type world;       // All world data 


/* Reset world params to inital values */
void world_reset( void )
{    
    world_free_animals();
    world_free_plants();
 
    // Inital cycle
    world.cycle = 0;
    
    // CLear animal pointer
    world.animals = NULL;     
 
    // No animals
    world.num_animals = 0;
 
    // No plants
    world.num_plants = 0;
 
    world.generation = 0;
    
    world.average = 0;
    world.score_sum = 0;
    
    world.wrap_around = false;
    world.graphics_on = true;
    world.running = true;
 
    world.updates_per_display = 1;
 
    // Add 50 animals 
    for (int i = 0; i < world.initial_animals; i++)
        world_add_animal();
        
    // Add 50 plants
    for (int i = 0; i < world.initial_plants; i++)
        add_random_plant();        
        
    // Calculate the mutate chance
    world.mutate_chance = 1.0 / (double) animal_get_chromosome_size(world.animals);        
    
}
/* End of reset world */



/* Inital world setup */
void world_setup(void )
{
    // Parameter configuration
    
    // Size in world space
    world.x_size = 10;
    world.y_size = 10;

    world.plant_size = 0.25;
    world.animal_size = 0.5;

    world.initial_plants = 10;
    world.initial_animals = 20;

    world.animal_neuron_width = 8;
    world.animal_neuron_depth = 2;

    // Very aggressive mutate rate
//    world.mutate_rate = 0.25;  // 0.005

    world.mutate_rate = 5.00;  // 0.005

    // Number of cycles before resampeling
    world.cycles_per_generation = 500;

    /* End of config parameters */


    write_console("world_setup\n");


    // Create brushs
    world.white_brush = CreateSolidBrush(RGB(255, 255, 255));
    world.black_brush = CreateSolidBrush(RGB(0, 0, 0));
    world.green_brush = CreateSolidBrush(RGB(0, 255, 0));
    world.blue_brush = CreateSolidBrush(RGB(0, 0, 255));
    world.red_brush = CreateSolidBrush(RGB(255, 0, 0));

    // Create pens
    world.white_pen = CreatePen(PS_SOLID, 1, RGB(255, 255, 255));
    world.black_pen = CreatePen(PS_SOLID, 1, RGB(0, 0, 0));
    world.green_pen = CreatePen(PS_SOLID, 2, RGB(0, 255, 0));

    world_reset();                               
}
/* End of void world_setup(void ) */


/* Free all world data */
void world_destroy( void )
{
    // Free the list of animals
    world_free_animals();
    world_free_plants();

    DeleteObject(world.black_brush);
    DeleteObject(world.white_brush);
    DeleteObject(world.green_brush);
    DeleteObject(world.blue_brush);
    DeleteObject(world.red_brush);
    
    DeleteObject(world.white_pen);
    DeleteObject(world.black_pen);
    DeleteObject(world.green_pen);
}
/* End of world_destroy */



/* Release animal memory */
void world_free_animals( void )
{
    animal_type *tmp;
        
    while (world.animals != NULL) 
    {
        tmp = world.animals->next;

        // Release neuron memory
        free_neuron_list(world.animals->neurons);

        // Release animal memory 
        free(world.animals);

        // Unlink from the list
        world.animals = tmp;
    }
    
    // Reset animal number
    world.num_animals = 0;
}
/* End of world_free_animals */


/* Free all plants */
void world_free_plants( void )
{
    plant_type *tmp;
        
    while (world.plants != NULL) 
    {
        tmp = world.plants->next;

        // Release animal memory 
        free(world.plants);

        world.plants = tmp; // Unlink
    }
    
    
    // Reset animal number
    world.num_plants = 0;
}    
/* End of world_free_plants */


/* Draw entire world */
void world_draw( HDC dc )
{
    if (!world.graphics_on) return;

    SelectObject(dc, world.white_pen);
    SelectObject(dc, world.white_brush);
      
    // Clear entire area
    Rectangle(dc, 0, 0, display_x, display_y);

    draw_plants(dc);
    
    SelectObject(dc, world.black_pen);
    SelectObject(dc, world.white_brush);

    // Draw all of the animals
    animal_type *a = world.animals;

    while (a != NULL)
    {       
        animal_draw(dc, a);
        a = a-> next;    
    }              
}
/* End of world_draw */


/* Update all world data */
void world_update( void )
{
    if (!world.running) return;

    // Update all of the animals
    animal_type *a = world.animals;

    while (a != NULL)
    {       
        animal_update(a);
        a = a-> next;    
    }              
        
    if( ++world.cycle == world.cycles_per_generation)
    {
        world_compute_generation();                   
    }
    
    updates++;
}
/* End of world_update */


/* Add an animal to the world */
void world_add_animal()
{
    animal_type *t = world.animals;
    
    world.animals = (animal_type *) malloc(sizeof(animal_type));

    world.animals->neurons = NULL;

    world.animals->x_size = world.animal_size;
    world.animals->y_size = world.animal_size;
    
    world.animals->next = t;
    
    animal_init(world.animals);
    
    world.num_animals ++;
}
/* End of world_add_animal */


/* Called one a second for displaying debug data */
void world_debug_update( void )
{
    if (world.animals == NULL) return;
}
/* End of world_debug */


/* Reset all adnimals for the next generation */
void world_reset_animals( void )
{
    animal_type *a = world.animals;
    
    while (a != NULL)
    {
        animal_reset(a);
        a = a->next;        
    }    
}
/* End of world_reset_animals */


/* User to compare the errors of two solutions */
int animal_compare(const void *s1, const void  *s2)
{
    animal_type *a1 = *(animal_type **)s1;
    animal_type *a2 = *(animal_type **)s2;

    if (a2->score > a1->score) return(1); 
    if (a2->score < a1->score) return(-1); 
    
    return(0);
}
/* End of error_compare */


/* Breed two solutions to form a third */
void combine_genome(double *out, double *ch1, double *ch2, int size)
{
    for (int i = 0; i < size; i++)
    {
        // Decide at random which term to use
        if (rand() % 2)
            out[i] = ch1[i];
        else
            out[i] = ch2[i];
    }
}
/* End of combine */


/* Display generation scores */
void world_show_scores( int total_plant, int total_wall )
{
    char buff[400];
    
    FILE *f = fopen("log.txt", "at");    
   
   double rel_score = total_plant / (double)total_wall;
   
   world.score_sum +=total_plant / (double)world.num_animals;

   world.average = world.score_sum / (double)(world.generation + 1);
   
    sprintf(buff, "Gen: %4d   Total Food: %5d  Total Wall: %6d   Ave food/animal: %5.3f\n", 
            world.generation, total_plant, total_wall, world.average);

    write_console(buff);
    fprintf(f, buff);
    
    animal_type * a = world.animals;
                
   // sprintf(buff, "Food    Wall    Score\n");
//    write_console(buff);
    //fprintf(f, buff);
                                                                   
   /* while (a != NULL)
    {8
        sprintf(buff, "%3d   %3d      %f\n", a->food_capture, a->wall_hit, a->score);
        
//        write_console(buff);        
        fprintf(f, buff);

        a = a->next;        
    }
*/

//    sprintf(buff, "");
//    write_console(buff);        
        
//    fprintf(f, buff);

    fclose(f);  
}
/* End of show scores */


/*  Adjust mutation rate based on number of good mutations */
void update_mutate_rate(int &good_mutate)
{
return;
    char buff[400];

    double rate_check = (good_mutate) / (10.0);

    // Apply faster rate if better than 1/5 of mutations were "good"
    if (rate_check > 1/5.0)
        world.mutate_rate /= 0.85;
    else
        world.mutate_rate *= 0.85;

    good_mutate = 0;

    sprintf(buff, "Mutate success rate: %f  Mutate rate: %f\n", rate_check , world.mutate_rate);

    write_console(buff);
}
/* End of pdate_mutate_rate */




/* Preform a mutation on a chomosome */
double mutate_value(double val, double chance, double rate)
{
    if (rand() < chance * RAND_MAX)
        return(val + f_rand(-rate, rate));
    else
        return(val);

}
/* End of animal_mutate_chromosome */



/* Cross over two animals and apply mutation */
void mate(animal_type *out, animal_type *a, animal_type *b,
              double rate, double chance)
{
    // Cross over an entire neuron at a time

    
    neuron_type *a_neuron = a->neurons;
    neuron_type *b_neuron = b->neurons;
    neuron_type *out_neuron = out->neurons;
    
    
    while (a_neuron != NULL && b_neuron != NULL)   
    {
        neuron_type *s;
        
        // Select either the a of b neuron
        if (rand() % 2)
            s = a_neuron;
        else
            s = b_neuron;
    
        out_neuron->bias = mutate_value(s->bias, rate, chance);
        
        for (int i = 0; i < s->num_sources; i++)
            out_neuron->source_scaler[i] = mutate_value( s->source_scaler[i], rate, chance);
        
        a_neuron = a_neuron->next;
        b_neuron = b_neuron->next;
        out_neuron  = out_neuron ->next;
    }
    
}
/* End of mate */


/* Compute next generation */
void world_compute_generation( void )
{
    if (world.animals == NULL) return;

    static double last_scores[50] = {0.0};
   
    static int good_mutate = 0;
        
    animal_type *sort[world.num_animals];       // List of all animal pointers for sort
    animal_type *a = world.animals;             // Point to first animal
        
    int high_plant = 0;
    int high_wall_hit = 0;
    
      
    static double last_score = 0;      
               
    // Fill in the pointer list
    for (int i = 0; i < world.num_animals; i++)
    {
        if (a->food_capture > high_plant)
            high_plant = a->food_capture; 

        if (a->wall_hit > high_wall_hit)
            high_wall_hit = a->wall_hit; 

        sort[i] = a;
        a = a->next;                       
    }
    
    
    // Normalize and compute scores
        
    a = world.animals;     

    for (int i = 0; i < world.num_animals; i++)
    {
        double norm_food = a->food_capture / (double)high_plant;
        double norm_wall = a->wall_hit / (double)high_wall_hit;
    
        // Compute score, hitting a wall is worth half scoring food
        a->score = norm_food;// - 0.5 * norm_wall;
    
       // a->score = a->food_capture - a->wall_hit;

        sort[i] = a;
        a = a->next;                       
    }   
    
                        
    // Sort animals, best score to worst
    qsort (sort, world.num_animals, sizeof(animal_type *), animal_compare);
    
    int total_plant = 0;  // Total score
    int total_wall = 0;  // Total score
        
    // Put best animal at root node 
    world.animals  = sort[0];
    
    a = world.animals;

    // Relink the linked list to represent the ordering
    for (int i = 1; i < world.num_animals; i++)
    {
        a->next = sort[i];     
     
        total_plant += a->food_capture;     
        total_wall += a->wall_hit;     
                 
        // Improved score in this position
//        if (a->score > last_scores[i])
//            good_mutate++;

        // Save old score
        last_scores[i] = a->score;

        a = a->next;                                     
    }

    a->next = NULL;    // Terminate new list

    if (total_plant > last_score)
        good_mutate++;

    last_score = total_plant;


    // Check the mutation rate after a set number of generations
    if (world.generation % 10 == 0 && world.generation != 0)
        update_mutate_rate(good_mutate);

    // Display scores for this generations
    world_show_scores(total_plant, total_wall);

    int partition = world.num_animals / 2;

    // Leave the top 50 percent of the solutions as they are
    // Combine the best solution with each solution folowwing is for the rest
    // Store the new solutions in the positions of the bottom 50 percent

    // Determine chromosome length

//    int size = animal_get_chromosome_size(world.animals);

    // Chromosome storage
//    double ch_best[size];
//    double ch_next_best[size];
    
//    double ch_parent[size];
//    double ch_out[size];
    
    // Export the chromosome for the bext animal
  //  animal_export_chromosome(sort[0], ch_best);
//    animal_export_chromosome(sort[1], ch_next_best);


    int best_index = 0;

    for (int i = 0; i < partition; i++)
    {
        // apply cross over to the beat and the one we are mating with
        // Store in first old one we are going to wipeout
       mate(sort[i+partition], sort[i+1], sort[best_index], 
                 world.mutate_chance, world.mutate_rate);  
                
        // CHeck source parent after half of the breeding
        if (i >= partition/2) 
            best_index = 1;
    }


  
    // Randomize the last animal
    //animal_randomize(sort[world.num_animals-1]);
    
    // Reset cycle counter
    world.cycle = 0;    
    
    // Put all animals back into starting position
    world_reset_animals();    
    
    world.generation++;        
}
/* End of world_compute_generation */
    
    
    
    
/* Save World to a file */
void world_save( void )
{
    OPENFILENAME ofname;
    char filename[300] = {0};
    char filter[] = "World files\0*.dat\0\0";             
             
    memset(&ofname, 0, sizeof(ofname));
             
    ofname.lStructSize = sizeof(ofname);
    ofname.lpstrFilter = filter; 
    ofname.lpstrFile = filename; 
    ofname.nMaxFile = 300; 
    ofname.lpstrFileTitle = NULL; 
//    ofname.nMaxFileTitle = 0; 
    ofname.lpstrInitialDir = "."; 

    ofname.lpstrDefExt = "dat";

    // Get filename from user
    if (!GetSaveFileName(&ofname))
        return;


//    strcat(filename, ".dat");

    FILE *f = fopen(filename, "wt");
    
    fprintf(f, "%f ", world.x_size);
    fprintf(f, "%f ", world.y_size);

    fprintf(f, "%f ", world.mutate_rate);
    fprintf(f, "%d ", world.cycles_per_generation);

    fprintf(f, "%d ", world.num_plants);
    fprintf(f, "%f ", world.plant_size);

    fprintf(f, "%d ", world.num_animals);
    fprintf(f, "%f ", world.animal_size);

    fprintf(f, "%d ", world.animal_neuron_width);
    fprintf(f, "%d\n", world.animal_neuron_depth);

    animal_type *a = world.animals;
    
    while(a != NULL)
    {
        neuron_type *n = a->neurons;
        fprintf(f, "%d\n", a->num_neurons);
        
        while (n != NULL)
        {
            fprintf(f, "%d  %f ", n->id, n->bias);
            fprintf(f, "%d ", n->num_sources);

            for (int i = 0; i < n->num_sources; i++)
                fprintf(f, "%f ", n->source_scaler[i]);                                        

            fprintf(f, "\n");        
            n = n->next;
        }    

        a = a->next;
    }
    
    fclose(f);
        
    char buff[400];
    sprintf(buff, "Saved to %s\n", filename);
    write_console(buff);
}
/* End of save world */
    
        
/* Load world from a file */
void world_load( void )
{
    OPENFILENAME ofname;

    char buff[400];    
    char filename[300] = {0};             
    char filter[] = "World files\0*.dat\0\0";
           
    memset(&ofname, 0, sizeof(ofname));
             
    ofname.lStructSize = sizeof(ofname);
    ofname.lpstrFilter = filter; 
    ofname.lpstrFile = filename; 
    ofname.nMaxFile = 300; 
    ofname.lpstrFileTitle = NULL; 
//    ofname.nMaxFileTitle = 300; 
    ofname.lpstrInitialDir = "."; 

    // Get filename from user
    if (!GetOpenFileName(&ofname))
        return;

    FILE *f = fopen(filename, "rt");

    fscanf(f, "%lf", &world.x_size);
    fscanf(f, "%lf", &world.y_size);

    fscanf(f, "%lf", &world.mutate_rate);
    fscanf(f, "%d", &world.cycles_per_generation);

    fscanf(f, "%d", &world.initial_plants);
    fscanf(f, "%lf", &world.plant_size);

    fscanf(f, "%d", &world.initial_animals);
    fscanf(f, "%lf", &world.animal_size);

    fscanf(f, "%d", &world.animal_neuron_width);
    fscanf(f, "%d\n", &world.animal_neuron_depth);

    sprintf(buff, "Adding %d animals and %d plants\n", 
        world.initial_animals, world.initial_plants);
    world_reset();



    


    write_console(buff);




    char b[400];
//    sprintf(b, "Num animals:  %d", world.num_animals);
//    MessageBox(NULL, b, "", 0);

    animal_type *a = world.animals;
        
    while(a != NULL)
    {
        neuron_type *n = a->neurons;
        fscanf(f, "%d\n", &a->num_neurons);
        
        while (n != NULL)
        {
            fscanf(f, "%d  %lf %d", &n->id, &n->bias, &n->num_sources);
   
            double fll;

            for (int i = 0; i < n->num_sources; i++)
                fscanf(f, "%lf", &n->source_scaler[i]);                                        
                    
            n = n->next;
        }    

        a = a->next;
    }
    
    fclose(f);
    
    //char buff[400];
    sprintf(buff, "Loaded from %s\n", filename);
    write_console(buff);
    
}
/* End of world load */
    
    
/* Transform from world space to screen space */
int get_screen_x( double x )
{
    return((int)(x / (double) world.x_size * display_x));
}
/* End of get_screen_x */


/* Transform from world space to screen space */
int get_screen_y( double y )
{
    return((int)(y / (double) world.y_size * display_y));
}
/* End of get_screen_y */


/* Return a bounting box in screen space for an object */
void get_bounding_rect(double x, double y, double x_size, double y_size, RECT &r)
{
    r.left = get_screen_x(x - x_size / 2);
    r.top = get_screen_y(y - y_size / 2);
    r.right = get_screen_x(x + x_size / 2);
    r.bottom = get_screen_y(y + y_size / 2);
}
/* End of get_bounding_rect */
    


    
    
    
    
    



